/*
** Module   : ChuiScreenDriver.java
** Abstract : A base class for ChUI screen driver.
**
** Copyright (c) 2013-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 MAG 20131114 Created initial version.
** 002 EVL 20131212 Added method to check if the driver is in batch mode.
** 003 EVL 20140115 Added method to check if the driver is in background mode.
** 004 HC  20140119 Implemented runtime support for the COLOR 
**                  protermcap-attribute.
** 005 MAG 20140326 Implements interactive client. 
** 006 MAG 20140507 Fix interactive client NPE when no arguments in command line. 
** 007 MAG 20140512 Add method to check screen interface type ChUI or GUI.
** 008 CA  20140805 Moved from ui.client.driver to new package, ui.client.chui.driver.
**                  Extracted common behavior to a base class.
** 009 CA  20141029 Added flag to identify if a client should be run in windows-compatible mode.
*/

package com.goldencode.p2j.ui.client.chui.driver;

import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.main.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.chui.driver.swing.*;
import com.goldencode.p2j.ui.client.driver.swing.*;
import com.goldencode.terminal.*;

/**
 * Base class for  ChUI driver.
 */
public abstract class ChuiScreenDriver<S extends ChuiSimulator>
extends GenericScreenDriver<BasePrimitives, ChuiOutputManager, S>
{
   /** Bogus terminal type name. */
   private String termType = "vt100";

   /**
    * Read configuration values from the bootstrap configuration and convert
    * these into an options instance.
    * <p>
    * The following parameters are supported:
    * <p>
    * <pre>
    * Category:Group:Key       Type
    * -----------------------  ---------------------------
    * client:chui:rows         integer
    * client:chui:columns      integer
    * client:chui:background   text which can be converted to a Color by Color.decode()
    * client:chui:foreground   text which can be converted to a Color by Color.decode()
    * client:chui:selection    text which can be converted to a Color by Color.decode()
    * client:chui:fontname     text containing a monospaced (fixed width) font family name
    * client:chui:fontsize     integer
    * </pre>
    *
    * @param    cfg
    *           The bootstrap configuration.
    *
    * @return   A new options instance with any defaults overridden as specified
    *           in the bootstrap configuration.
    */
   public static TerminalOptions convertToOptions(BootstrapConfig cfg)
   {
      TerminalOptions options = new TerminalOptions();
      
      String next;
      
      if ((next = cfg.getString("client", "chui", "rows", null)) != null)
      {
         options.rows = parseInt(next, options.rows);
      }
      
      if ((next = cfg.getString("client", "chui", "columns", null)) != null)
      {
         options.columns = parseInt(next, options.columns);
      }
      
      if ((next = cfg.getString("client", "chui", "background", null)) != null)
      {
         options.background = parseColor(next, options.background);
      }
      
      if ((next = cfg.getString("client", "chui", "foreground", null)) != null)
      {
         options.foreground = parseColor(next, options.foreground);
      }
      
      if ((next = cfg.getString("client", "chui", "selection", null)) != null)
      {
         options.selection = parseColor(next, options.selection);
      }
      
      if ((next = cfg.getString("client", "chui", "fontname", null)) != null)
      {
         options.fontname = next;
      }
      
      if ((next = cfg.getString("client", "chui", "fontsize", null)) != null)
      {
         options.fontsize = parseInt(next, options.fontsize);
      }
      
      if ((next = cfg.getString("client", "chui", "windows", null)) != null)
      {
         options.underWindows = "true".equalsIgnoreCase(next);
      }
      
      return options;
   }

   /**
    * Attempt to convert the given string into an integer using
    * <code>Integer.parseInt</code>.
    *
    * @param    spec
    *           String encoded number specification.
    * @param    def
    *           Default value to return if the spec cannot be parsed properly.
    *
    * @return   The parsed number or the default if there is any parsing
    *           problem.
    */
   private static int parseInt(String spec, int def)
   {
      int val = def;
      
      try
      {
         val = Integer.parseInt(spec);
      }
      
      catch (NumberFormatException nfe)
      {
         // ignore
      }
      
      return val;
   }

   /**
    * Attempt to convert the given string into a color using
    * <code>Color.decode</code>.
    *
    * @param    spec
    *           String encoded color specification.
    * @param    def
    *           Default value to return if the spec cannot be parsed properly.
    *
    * @return   The parsed color or the default if the specification was
    *           invalid.
    */
   private static java.awt.Color parseColor(String spec, java.awt.Color def)
   {
      java.awt.Color color = def;
      
      try
      {
         color = java.awt.Color.decode(spec);
      }
      
      catch (NumberFormatException nfe)
      {
         // ignore
      }
      
      return color;
   }

   /**
    * Get the {@link OutputManager} instance associated with this driver.
    * 
    * @param    cfg
    *           The initial configuration for this driver.
    * 
    * @return   A {@link ChuiOutputManager} instance.
    */
   @Override
   public ChuiOutputManager createOutputManager(BootstrapConfig cfg)
   {
      return new ChuiOutputManager(this);
   }
   
   /**
    * Driver initialization.
    */
   @Override
   public void init()
   {
      setPrimitives(new ChuiPrimitives((ChuiSimulator) sim));
   }
   
   /**
    * Perform driver shutdown.
    */
   public void shutdown()
   {
      // no-op
   }

   /**
    * Get driver-specific widget factory.
    * 
    * @return  reference to widget factory.
    */
   @Override
   public ChuiWidgetFactory getFactory()
   {
      if (factory == null)
      {
         factory = new ChuiWidgetFactory();
      }
      
      return (ChuiWidgetFactory) factory;
   }
   
   /**
    * Get the default WINDOW-SYSTEM used by this driver. See {@link ClientParameters} for the list
    * of possible values.
    * 
    * @return   See above.
    */
   @Override
   public String getWindowSystem()
   {
      return ClientParameters.WINDOW_SYSTEM_TTY;
   }
   
   /**
    * Get screen interface type.
    * 
    * @return  <code>true</code> is ChUI <code>false</code> is GUI
    */
   @Override
   public boolean isChui()
   {
      return true;
   }

   /**
    * Change terminal type to specified one. The terminal-based implementation
    * should try to change terminal type to specified one. It is assumed that 
    * passed terminal string is specific to particular driver implementation 
    * and recognized by it. 
    * 
    * @param   name
    */
   @Override
   public void setTerminalType(String name)
   {
      termType = name;
   }

   /**
    * Obtain system/implementation specific terminal type string.
    * <p>
    * At present this is used to implement 4GL TERMINAL function and do some 
    * terminal-specific workarounds. At present workarounds exist for "vt320" 
    * terminal type. If implementation requires no special workarounds, then 
    * it should avoid returning "vt320" as terminal type string.  
    * 
    * @return  terminal type string.
    */
   @Override
   public String terminalType()
   {
      return termType;
   }

   /**
    * Set cursor status (ON/OFF).
    * <p>
    * The implementation of this method should show/hide actual screen 
    * cursor visible to user. Remaining code assumes that actual cursor status
    * visible to user is properly changed upon exit from this method.
    * 
    * @param   on
    *          <code>true</code> - show cursor, <code>false</code> - hide cursor.
    * @return  <code>true</code> if operation was successful.
    * 
    * @throws  IllegalStateException
    */
   @Override
   public boolean setCursorStatus(boolean on)
   throws IllegalStateException
   {
      return sim.setCursorStatus(on);
   }

   /**
    * Create a shell emulator and execute the given command.
    * 
    * @param    cmdList
    *           The command.
    * @param    overlay
    *           Flag identifying if the emulator window will overlay the current window or will
    *           open in a new window.
    *           
    * @return   A VT100 emulator executing the given command.
    */
   @Override
   protected VT100Emulator createShellEmulator(String[] cmdList, int overlay)
   {
      // Renderer
      VT100SwingRenderer term = new VT100SwingRenderer(sim);

      // Keyboard
      VT100SwingKeyboard key = new VT100SwingKeyboard();
      sim.setVT100Keyboard(key);

      return new VT100Emulator(term, key, cmdList, overlay);
   }
}
