/*
** Module   : AbstractSimulator.java
** Abstract : Base class for all swing simulators.
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140805 Created initial version.  Extracted from code in previous ChuiSimulator class.
** 002 CA  20141029 Each simulator needs to receive the swing keyboard reader implementation in 
**                  use.  It's the callers responsibility to ensure that the same instance is used
**                  for all simulators, if the client is ran in a multi-window mode (i.e. GUI).
*/

package com.goldencode.p2j.ui.client.driver.swing;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.awt.print.*;
import java.util.*;

import javax.swing.*;

import com.goldencode.p2j.ui.client.swing.*;
import com.goldencode.terminal.*;

/**
 * Base class for all swing simulators. 
 */
@SuppressWarnings("serial")
public abstract class AbstractSimulator
extends JComponent
implements Printable
{
   /** Lock object for synchronizing configuration usage and changes. */
   protected Object cfgLock = new Object();
   
   /** Lock object for synchronizing selection processing. */
   protected Object selLock = new Object();
   
   /** Lock object for synchronizing data and cursor usage and changes. */
   protected Object dataLock = new Object();

   /** Helper to map color specifications to the font needed for drawing. */
   protected ColorMapper colorMap = null;
   
   /** Flag that controls if the cursor is visible or not. */
   protected boolean showCursor = true;
   
   /** Containing frame (may be an applet or an application). */
   protected Frame frame = null;
   
   /** Key processing for VT-100 emulator. */
   private VT100AbstractKeyboard keyboardVT100 = null;

   /** The key reader associated with this simulator. */
   protected final SwingKeyboardReader keyReader;
   
   /**
    * Default constructor.
    */
   public AbstractSimulator()
   {
      this(null);
   }
   
   /**
    * Create a new simulator and set its key listener.
    * 
    * @param    keyReader
    *           The {@link SwingKeyboardReader key reader} which listens for incoming keys.
    */
   public AbstractSimulator(SwingKeyboardReader keyReader)
   {
      this.keyReader = keyReader;
   }

   /**
    * Create a new instance with the container and frame given.  When this method is done, all 
    * listeners and key processing will be initialized.
    * <p>
    * For applications, the frame and the container are usually the same
    * object (of type <code>JFrame</code>). For applets, these are different
    * instances since a <code>JApplet</code> will be the container and the
    * frame will be a parent container of the <code>JApplet</code>.
    *
    * @param    frame
    *           The containing frame. Needed for key event processing
    *           and for the processing of modal dialog boxes.
    * @param    container
    *           The immediate container component, which can be called to
    *           force a resizing to match a changed simulator size.
    * @param    keyReader
    *           The {@link SwingKeyboardReader key reader} which listens for incoming keys.
    */
   public AbstractSimulator(Frame frame, ContainerOps container, SwingKeyboardReader keyReader)
   {
      this(keyReader);
      
      this.frame = frame;

      if (keyReader == null)
      {
         throw new NullPointerException("The key reader can not be null!");
      }
      
      // listen for keystrokes
      frame.addKeyListener(keyReader);
      
      // disable focus processing on TAB and SHIFT-TAB, otherwise those
      // keys are not ever available for consumption
      frame.setFocusTraversalKeys(KeyboardFocusManager.FORWARD_TRAVERSAL_KEYS,
                                  new HashSet<AWTKeyStroke>());
      frame.setFocusTraversalKeys(KeyboardFocusManager.BACKWARD_TRAVERSAL_KEYS,
                                  new HashSet<AWTKeyStroke>());
   }

   /**
    * Overwrite the entire 2-dimensional output buffer with the space character
    * to be rendered with the <code>Color.NORMAL</code> color. Sets the cursor
    * Y and X positions to 0.
    */
   public abstract void clear();

   /**
    * Set the cursor visibility.
    * <p>
    * By default it returns {@code false} and this impl is a no-op.
    * 
    * @param    on
    *           <code>true</code> to make the cursor visible (subsequent repaint
    *           operations will then cause the cursor to be drawn).
    *
    * @return   The previous cursor visiblity state.
    */
   public boolean setCursorStatus(boolean on)
   throws IllegalStateException
   {
      return false;
   }
   
   /**
    * Capture and return the terminal screen as an image.
    *
    * @return   An image representing the terminal screen that is being
    *           displayed.
    */
   public Image capture()
   {
      BufferedImage image = new BufferedImage(getWidth(),
                                              getHeight(),
                                              BufferedImage.TYPE_4BYTE_ABGR);
      
      printAll(image.createGraphics());
      
      return image;
   }
   
   /**
    * Create a new printer job, open the job dialog for the user and if
    * printing is not canceled by the user, print the current terminal screen
    * to the chosen printer, with the configured job settings.
    */
   public void print()
   {
      PrinterJob job = PrinterJob.getPrinterJob();
      job.setPrintable(this);
      
      if (job.printDialog())
      {
         try
         {
            job.print();
         }
         
         catch (PrinterException pe)
         {
            // ignore
         }
      }      
   }
   
   /**
    * Low level printing worker that is called back by the Java printer job
    * in order to generate the print output. This will only ever print a single
    * page of output. It does not scale the print output to the page, so it
    * is possible that the output will be cut off to the right and on the
    * bottom.
    *
    * @param    g
    *           The graphics instance into which to draw.
    * @param    pf
    *           Details of the printer page being processed.
    * @param    page
    *           The page number being requested. Only the 0 page will ever
    *           be honored.
    *
    * @return   <code>PAGE_EXISTS</code> if the page was printed or
    *           <code>NO_SUCH_PAGE</code> if any page number other than 0
    *           is requested.
    */
   public int print(Graphics g, PageFormat pf, int page)
   throws PrinterException
   {
      if (page > 0)
      {
         return NO_SUCH_PAGE;
      }

      Graphics2D graphics = (Graphics2D) g;
      graphics.translate(pf.getImageableX(), pf.getImageableY());

      printAll(graphics);

      return PAGE_EXISTS;
   }   
   
   /**
    * Cause the outer frame to be disposed, exiting the application or applet.
    * This is important since <code>System.exit()</code> cannot be used for
    * applets.
    */
   public void quit()
   {
      // protect against multiple calls in the same session
      if (frame.isDisplayable())
         frame.dispose();
   }
   
   /**
    * Read the next key from the FIFO key input buffer.  This will block until
    * a key is available.
    *
    * @return   The next key in the buffer.
    */
   public int readKey()
   {
      return keyReader.read();
   }
   
   /**
    * Insert the given characters into the keyboard input buffer (at the end, in FIFO order).
    *
    * @param    data
    *           Characters to insert into the "keyboard" buffer.
    */
   public void paste(String data)
   {
      keyReader.insert(data);
   }
   
   /**
    * Switch between P2J client and interactive client.
    * 
    * @param   toVT100
    *          If <code>true</code> show interactive process screen otherwise P2J client screen.
    */
   public void swapKeyboards(boolean toVT100)
   {
      JFrame app = (JFrame) frame;
            
      if (toVT100)
      {
         app.addKeyListener((KeyListener) keyboardVT100);
         app.removeKeyListener(keyReader);
      }
      else
      {
         app.addKeyListener(keyReader);
         app.removeKeyListener((KeyListener) keyboardVT100);
      }
   }

   /**
    * Set keyboard listener for VT100 emulator.
    *  
    * @param   keyboard
    *          A VT100 keyboard instance.
    */
   public void setVT100Keyboard(VT100AbstractKeyboard keyboard)
   {
      keyboardVT100 = keyboard;
   }

   /**
    * Check if this simulator is for a web client.
    * 
    * @return   By default, return {@code false}.
    */
   public boolean isWeb()
   {
      return false;
   }

   /**
    * Handle JS key codes.
    * 
    * @param   keyCode
    *          Key code.
    * @param   charCode
    *          Character code
    */   
   public void injectKey(int keyCode, char charCode)
   {
      keyboardVT100.injectKey(keyCode, charCode);
   }
}
