/*
** Module   : GuiSimulator.java
** Abstract : GUI renderer.
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 MAG 20140508 First version.
** 002 CA  20140711 Enhanced font support - removed fontSize.
** 003 CA  20140806 Added mouse support and other misc APIs related to window state.
** 004 MAG 20140912 Added fill polygon.
** 005 HC  20140911 GUI coordinate support - data types refactoring.
** 006 MAG 20140924 Create graphical object when buffered image is created.
** 007 CA  20141021 Added notification of OS window events.  Added support for batch drawing: the 
**                  current AWT repaint processing will wait until the last paint primitive is
**                  received, before returning the control to the AWT event processing thread, to
**                  avoid flickering.  Do not allow window maximize via OS means.
**                  Added implementation for new paint primitives: SET_LOCATION, RESIZE_WINDOW.
** 008 CA  20141029 Each simulator needs to receive the swing keyboard reader implementation in 
**                  use.  It's the callers responsibility to ensure that the same instance is used
**                  for all simulators, if the client is ran in a multi-window mode (i.e. GUI).
*/

package com.goldencode.p2j.ui.client.gui.driver.swing;

import java.awt.*;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.event.*;
import java.awt.image.*;
import java.util.*;
import java.util.concurrent.*;

import javax.swing.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.*;
import com.goldencode.p2j.ui.client.driver.swing.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.WindowEvent;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.swing.*;

/**
 * The GUI renderer. Because the paint should be permanent a BufferedImage is used to draw
 * and keep the last screen image. A pool of drawing operations is used to queue drawing 
 * primitives. The operations are enqueued from poll one by one and draw on the buffered image. 
 * When pool is empty the screen is updated with the image from the stored buffered image.  
 */
@SuppressWarnings("serial")
public class GuiSimulator 
extends AbstractSimulator
{
   /** Rectangle instance to identify a "NO-CLIP" state. */
   private static final java.awt.Rectangle NO_CLIP = new java.awt.Rectangle();

   /** Screen buffer */
   private BufferedImage screenBuffer;
   
   /** Graphical object */
   private Graphics2D g2;

   /** Draw operations pools */
   private Deque<PaintStructure> graphicPool = new ConcurrentLinkedDeque<>();

   /** The window ID associated with this simulator. */
   private int windowId;
   
   /** A stack of clipping rectangles. */
   private Deque<java.awt.Rectangle> clippings = new ConcurrentLinkedDeque<>();

   /** The registered mouse handler. */
   private final MouseHandler mouseHandler;
   
   /** The bitmap associated with this simulator. */
   private ScreenBitmap bitmap;

   /**
    * Constructor.
    * 
    * @param   frame
    *          Application container.
    * @param   options
    *          GUI options.
    * @param   keyReader
    *          The {@link SwingKeyboardReader key reader} which listens for incoming keys.
    */
   public GuiSimulator(JFrame frame, GuiOptions options, SwingKeyboardReader keyReader)
   {
      super(frame, null, keyReader);

      frame.setUndecorated(true);

      // register listeners for OS window events
      final ThinClient tc = ThinClient.getInstance();
      frame.addWindowListener(new WindowAdapter()
      {
         @Override
         public void windowIconified(java.awt.event.WindowEvent e)
         {
            tc.eventBracket(new Runnable()
            {
               @Override
               public void run()
               {
                  WindowGuiImpl window = (WindowGuiImpl) tc.getWidget(windowId);
                  EventManager.postEvent(new WindowEvent(window, Keyboard.SE_WINDOW_MINIMIZED));
               }
            });
         }
         
         @Override
         public void windowDeiconified(java.awt.event.WindowEvent e)
         {
            tc.eventBracket(new Runnable()
            {
               @Override
               public void run()
               {
                  WindowGuiImpl window = (WindowGuiImpl) tc.getWidget(windowId);
                  EventManager.postEvent(new WindowEvent(window, Keyboard.SE_WINDOW_RESTORED));
               }
            });
         }
      });

      frame.addWindowStateListener(new WindowStateListener()
      {
         public void windowStateChanged(java.awt.event.WindowEvent e)
         {
            if ((e.getNewState() & Frame.MAXIMIZED_BOTH) == Frame.MAXIMIZED_BOTH)
            {
               resizeWindow(getWidth(), getHeight());
            }
         }
      });
      
      // Initialize screen bitmap
      Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
      screenBuffer = new BufferedImage(screenSize.width, 
                                       screenSize.height, 
                                       BufferedImage.TYPE_INT_ARGB);
      g2 = screenBuffer.createGraphics();
   
      // setup mouse listeners
      mouseHandler = new MouseHandler(this);
      addMouseListener(mouseHandler);
      addMouseMotionListener(mouseHandler);
      
      // TODO: disable the popup of the window actions (min/max/close/etc) via key? 
   }

   /**
    * Enable or disable capture of mouse events.
    * 
    * @param    state
    *           {@code true} to capture mouse events, {@code false} to ignore them.
    *           
    * @return   The previous state.
    */
   public boolean captureMouseEvents(boolean state)
   {
      return mouseHandler.captureMouseEvents(state);
   }
   
   /**
    * Hide or show the window, depending on the {@code visible} flag.
    * 
    * @param    visible
    *           The window visibility state.
    */
   public void setWindowVisible(boolean visible)
   {
      frame.pack();
      frame.setVisible(visible);
   }

   /**
    * Set the window ID for this simulator.
    * 
    * @param   windowId
    *          The window ID.
    */
   public void setWindowId(int windowId)
   {
      this.windowId = windowId;
   }

   /**
    * Get the window ID for this simulator.
    */
   public int getWindowId()
   {
      return windowId;
   }

   /**
    * Paint the entire component to the given graphics instance, using
    * previously set state.  Depending on the repaint request, the graphics
    * object's clipping region may have been set to a subset of the full
    * rectangle. This method makes no such distinction, it just blindly
    * draws everything and the stuff outside of the clipping region never
    * gets sent to the hardware.  This is the core where all drawing occurs
    * and it draws the screen text (including switching out the fonts per
    * character as needed to get the right colors/attributes), the current
    * cursor (if it is visible) and the any selection box (when a selection
    * is active).
    *
    * @param    graphics
    *           The object into which to draw.
    */
   @Override
   public void paintComponent(Graphics graphics)
   {
      super.paintComponent(graphics);

      // draw
      drawOnScreenBitmap(graphics);
   }  

   /**
    * Add a new <code>PaintStructure</code> to the drawing pool.
    * The item is enqueued at the end of the pool.
    * 
    * @param   ps
    *          A new PaintStructure instance.          
    */      
   public void offer(PaintStructure ps)
   {
      // to prevent deadlocks, synchronize on the frame's layout lock
      synchronized (getTreeLock())
      {
         graphicPool.offerLast(ps);

         getTreeLock().notify();
      }
   }

   /**
    * Resize window.
    * 
    * @param   width
    *          new width.
    * @param   height
    *          new height
    */
   public void resizeWindow(int width, int height)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.RESIZE_WINDOW);
      ps.width = width;
      ps.height = height;
      
      offer(ps);
      repaint();

      this.bitmap.setRows(height);
      this.bitmap.setColumns(width);
   }

   /**
    * Set window location.
    * 
    * @param   left
    *          left coordinate
    * @param   top
    *          top coordinate
    */
   public void setWindowLocation(int left, int top)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.SET_LOCATION);
      ps.x = left;
      ps.y = top;
      
      offer(ps);
      repaint();
   }
   
   /**
    * Get the window location on screen.
    * 
    * @return   See above.
    */
   public NativePoint getWindowLocation()
   {
      java.awt.Point p = frame.getLocation();
      
      return new NativePoint(p.x, p.y);
   }

   /**
    * Iconify the window.
    */
   public void iconifyWindow()
   {
      frame.setState(Frame.ICONIFIED);
   }

   /**
    * Limit the available drawing area to the specified boundaries.
    * 
    * @param    bounds
    *           The bounds for the drawing area.
    */
   public void setDrawingArea(NativeRectangle bounds)
   {
      PaintStructure ps = new PaintStructure(bounds == null ? PaintPrimitives.NO_CLIP  
                                                            : PaintPrimitives.CLIP);
      if (bounds != null)
      {
         ps.x = bounds.left();
         ps.y = bounds.top();
         ps.width = bounds.right() - bounds.left();
         ps.height = bounds.bottom() - bounds.top();
      }
      
      offer(ps);
   }

   /**
    * Get the screen width, in character units.
    * 
    * @return    See above.
    */
   public int screenWidth()
   {
      return bitmap.cols();
   }

   /**
    * Get the screen height, in character units.
    * 
    * @return    See above.
    */
   public int screenHeight()
   {
      return bitmap.rows();
   }

   /**
    * Set the active screen bitmap for the current window.
    * 
    * @param    bitmap
    *           The screen bitmap.
    */
   public void setScreenBitmap(ScreenBitmap bitmap)
   {
      this.bitmap = bitmap;
   }

   /**
    * Get the screen bitmap for the current window.
    * 
    * @return   See above.
    */
   public ScreenBitmap getScreenBitmap()
   {
      return bitmap;
   }

   /**
    * Start a batch drawing routine.
    */
   public void startBatch()
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.START_BATCH);
      offer(ps);
   }

   /**
    * End a batch drawing routine.
    */
   public void endBatch()
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.END_BATCH);
      offer(ps);
      
      repaint();
   }
   
   /**
    * Clear the entire screen, with the default window color.
    */
   @Override
   public void clear()
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.CLEAR);
      ps.x = 0;
      ps.y = 0;
      ps.width = getWidth();
      ps.height = getHeight();
      ps.color = null; // TODO: color???
      
      offer(ps);
   }

   /**
    * Clear specified area at the screen.
    * 
    * @param   left
    *          Left coordinate of the area.
    * @param   top
    *          Top coordinate of the area.
    * @param   right
    *          Right coordinate of the area.
    * @param   bottom
    *          Bottom coordinate of the area.
    * @param   color
    *          Color.
    */
   public void clearArea(int left, int top, int right, int bottom, com.goldencode.p2j.ui.Color color)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.CLEAR_AREA);
      ps.x = left;
      ps.y = right;
      ps.width = right - left;
      ps.height = bottom - top;
      ps.color = null; // TODO: color???
      
      offer(ps);
   }

   /**
    * Update screen. Iterate over the pool and draw.
    * 
    * @param   graphics
    *          Graphical context.
    */
   private void drawOnScreenBitmap(Graphics graphics)
   {
      // ensure the drawing will be performed by only one thread at a time

      synchronized (screenBuffer)
      {
         int waitForBatch = 0;
         while (!graphicPool.isEmpty() || waitForBatch != 0)
         {
            // if a batch was started, wait for it to end.
            synchronized (getTreeLock())
            {
               if (graphicPool.isEmpty() && waitForBatch != 0)
               {
                  try
                  {
                     getTreeLock().wait();
                  }
                  catch (InterruptedException e)
                  {
                     // terminate painting
                     break;
                  }
               }
            }
            
            PaintStructure ps = graphicPool.pollFirst();

            if (ps != null)
            {
               switch (ps.id)
               {
                  case FLUSH:
                     graphics.drawImage(screenBuffer, 0, 0, this);
                     break;
                  
                  case START_BATCH:
                     waitForBatch++;
                     break;

                  case END_BATCH:
                     waitForBatch--;
                     break;
                     
                  default:
                     draw(g2, ps);
                     break;
               }
            }
         }
   
         graphics.drawImage(screenBuffer, 0, 0, this);
      }
   }

   /**
    * Draw using a paint operation.
    * 
    * @param   g2
    *          Graphical context.
    * @param   ps
    *          Paint operation parameters.
    */
   private void draw(Graphics2D g2, PaintStructure ps)
   {
      switch (ps.id)
      {
         case DRAW_STRING:
            g2.drawString(ps.text, ps.x, ps.y);
            break;
         case DRAW_LINE:
            g2.drawLine(ps.x, ps.y, ps.width, ps.height);
            break;
         case DRAW_RECT:
            g2.drawRect(ps.x, ps.y, ps.width, ps.height);
            break;
         case DRAW_IMAGE:
            g2.drawImage(ps.img, ps.x, ps.y, ps.width, ps.height, this);
            break;
         case FILL_RECT:
            g2.fillRect(ps.x, ps.y, ps.width, ps.height);
            break;
         case FILL_POLYGON:
            g2.fillPolygon(ps.xPoints, ps.yPoints, ps.nPoints);
            break;
         case DRAW_3D_RECT:
            g2.draw3DRect(ps.x, ps.y, ps.width, ps.height, ps.raised);
            break;
         case FILL_3D_RECT:
            g2.fill3DRect(ps.x, ps.y, ps.width, ps.height, ps.raised);
            break;
         case SET_COLOR:
            g2.setColor(ps.color);
            break;
         case SET_FONT:
            g2.setFont(ps.font);
            break;
         case SET_FONT_STYLE:
            g2.setFont(g2.getFont().deriveFont(ps.x));
            break;
         case TRANSLATE:
            g2.translate(ps.x, ps.y);
            break;
         case CLEAR_AREA:
            if (ps.color != null)
            {
               g2.setColor(ps.color);
            }
            g2.fillRect(ps.x, ps.y, ps.width, ps.height);
            break;
         case CLEAR:
            if (ps.color != null)
            {
               g2.setColor(ps.color);
            }
            g2.fillRect(ps.x, ps.y, ps.width, ps.height);
            break;
         case CLIP:
            java.awt.Rectangle prevClip = g2.getClipBounds();
            
            clippings.push(prevClip == null ? NO_CLIP : prevClip);
            
            // debugging aid: enable these lines to outline the clipped area
            if (false)
            {
               g2.setColor(Color.green);
               g2.drawRect(ps.x, ps.y, ps.width, ps.height);
            }
            
            g2.setClip(ps.x, ps.y, ps.width, ps.height);
            break;
         case NO_CLIP:
            java.awt.Rectangle clip = clippings.isEmpty() ? null : clippings.pop();
            clip = (clip == NO_CLIP ? null : clip);
            g2.setClip(clip);
            break;
         case SET_LOCATION:
            frame.setLocation(ps.x, ps.y);
            break;
         case RESIZE_WINDOW:
            Dimension sz = new Dimension(ps.width, ps.height);

            setPreferredSize(sz);
            setMinimumSize(sz);
            setMaximumSize(sz);
            
            frame.setSize(sz);
            setSize(sz);
            
            frame.pack();
            break;
         default:
            break;
      }      
   }
}
